#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example script which prints all note on events on MIDI channel 0."""

import sys

from os.path import abspath, dirname, join

from miditk.smf import MidiFileReader, NullMidiEventHandler


class ChannelZeroPrinter(NullMidiEventHandler):
    """This prints all note on/off events on MIDI channel 0."""

    def note_on(self, channel=0, note=0x40, velocity=0x40):
        if channel == 0:
            print(channel, note, velocity, self.relative_time)

    note_off = note_on


try:
    infile = sys.argv[1]
except IndexError:
    infile = join(dirname(dirname(abspath(__file__))), 'tests', 'testdata',
                  'minimal-cubase-type1.mid')

# create the reader and the event handler
midiin = MidiFileReader(infile, handler=ChannelZeroPrinter())
# do parsing
midiin.read()



# -*- coding: utf-8 -*-
#
# miditk/smf/sequence.py
#
"""A MIDI stream event handler class to convert an SMF file into a container
object for a sequence of MIDI time-stamped events."""

from __future__ import absolute_import, print_function, unicode_literals

# standard library imports
import logging
from itertools import groupby
from operator import attrgetter

# package-specific imports
from ..common.constants import (END_OF_EXCLUSIVE, END_OF_TRACK, KEY_SIGNATURE, SEQUENCE_NAME,
                                SYSTEM_EXCLUSIVE, TEMPO, TIME_SIGNATURE)
from .api import BaseMidiEventHandler
from .converters import tointseq
from .reader import MidiFileReader


__all__ = ('MidiSequence', 'ObjectMidiEventHandler')
log = logging.getLogger(__name__)


class MidiSequence(list):
    """Container for a sequence of time-stamped MIDI events."""

    def __init__(self, *args, **kwargs):
        super(MidiSequence, self).__init__(*args, **kwargs)
        self.tracks = {}

    def __repr__(self):
        title = getattr(self, 'sequence_name', '')
        return "<MidiSequence tracks=%i events=%i, title='%s'>" % (
            self.num_tracks, len(self), title)

    def dump_events(self):
        for i, ev in enumerate(sorted(self, key=attrgetter('time', 'track'))):
            if ev.channel is not None:
                print("%03i: Channel event - type: %02XH, channel: %X, data: %r [%.2f]" %
                      (i, ev.type, ev.channel, tointseq(ev.data), ev.time))
            elif ev.meta_type:
                print('%03i: Meta event - type: %02XH, data: %r [%.2f]' % (
                      i, ev.meta_type, tointseq(ev.data), ev.time))
            elif ev.type == SYSTEM_EXCLUSIVE:
                print('%03i: Sysex_event - size: %i [%.2f]' % (
                      i, len(ev.data) + 1, ev.time))

    def events_by_ticks(self):
        return groupby(sorted(self, key=attrgetter('ticks', 'track')),
                       key=attrgetter('ticks'))

    def events_by_time(self):
        return groupby(sorted(self, key=attrgetter('time', 'track')),
                       key=attrgetter('time'))

    def events_by_track(self):
        return groupby(sorted(self, key=attrgetter('track', 'time')),
                       key=attrgetter('track'))

    def channel_message_events(self, track=None):
        for event in sorted(self, key=attrgetter('track', 'time')):
            if event.channel and (track is None or event.track == track):
                yield event

    def meta_events(self, track=None):
        for event in sorted(self, key=attrgetter('track', 'time')):
            if event.meta_type and (track is None or event.track == track):
                yield event

    def sysex_events(self, track=None):
        for event in sorted(self, key=attrgetter('time', 'track')):
            if (event.type == SYSTEM_EXCLUSIVE and
                    (track is None or event.track == track)):
                yield event

    @classmethod
    def fromfile(cls, infile, debug=False):
        sequence = cls()
        midiin = MidiFileReader(infile, ObjectMidiEventHandler(sequence, debug=debug))
        midiin.read()
        return sequence


class ObjectMidiEventHandler(BaseMidiEventHandler):
    """Collects all information from a MIDI file in a container object.

    On class instantiation you need to pass an instance to the constructor,
    which supports the Python list interface and attribute assignment. If
    None is passed (the default), a ``MidiSequence`` instance will be created
    and used instead.

    """
    meta_events = ['copyright', 'cuepoint', 'device_name', 'end_of_track',
                   'instrument_name', 'key_signature', 'lyric', 'marker',
                   'midi_ch_prefix', 'midi_port', 'program_name', 'sequence_name',
                   'sequence_number', 'sequencer_specific', 'smtp_offset', 'tempo',
                   'text', 'time_signature']
    channel_events = ['aftertouch', 'channel_pressure', 'controller_change',
                      'note_off', 'note_on', 'pitch_bend', 'program_change']

    def __init__(self, instance=None, debug=False):
        if instance is None:
            instance = MidiSequence()
        BaseMidiEventHandler.__init__(self)
        self._instance = instance
        self.current_time = 0.0
        self.current_tempo = 500000
        self.tick_division = 96
        self.running_status = None
        self._sysex_continuation = False
        self._sysex_buffer = None
        self.debug = debug

    def __getattr__(self, name):
        if name in self.meta_events:
            return self.meta_event
        elif name in self.channel_events:
            return self.channel_message
        else:
            raise AttributeError(name)

    #########################
    ## time handling event handlers.
    ## They should be overwritten with care

    def update_ticks(self, ticks=0, relative=True):
        """Update the global time."""
        BaseMidiEventHandler.update_ticks(self, ticks, relative)
        self.current_time += ticks * (self.current_tempo / 1000.0 /
                                      self._instance.tick_division)

        if self.debug:
            log.debug("Timer update: %s ticks", ticks)

    def add_event(self, event, track=None):
        """Add given event to the container object.

        Takes care of adding the current sequence time to the object.

        """
        event.time = self.current_time
        event.ticks = self.absolute_time
        event.relative_time = self.relative_time
        event.track = self.current_track if track is None else track
        self._instance.append(event)

    #########################
    ## Parsing (non-midi) events

    ## MIDI file header handler

    def header(self, format=0, num_tracks=1, tick_division=96, metrical=True,
               fps=0xE7, frame_resolution=0x28):
        """Handle MIDI file header."""
        for key, value in locals().items():
            if key != 'self' and not key.startswith('_'):
                setattr(self._instance, key, value)
        if metrical:
            log.debug("MIDI file format: %s, no. of tracks: %i, "
                      "tick division: %i ppqn", format, num_tracks, tick_division)
        else:
            log.debug("MIDI file format: %s, no. of tracks: %i, fps: %i, "
                      "resolution: %i", format, num_tracks, fps, frame_resolution)

    def eof(self):
        """Handle end of MIDI file."""
        self.current_track = None

        if self.debug:
            log.debug("End of file reached.")

    def start_of_track(self, track):
        """Handle start of new track."""
        if self.debug:
            log.debug("Start of track #%i.", track)

        self.current_track = track
        self.current_time = 0.0

    #############################
    ## Channel events

    def channel_message_event(self, event, *args):
        """Handle channel messages events."""
        if self.debug:
            log.debug("Channel message event: %s", event)
        self.add_event(event)

    ###############
    ## sysex event

    def invalid_event(self, event):
        """Handle invalid events."""
        log.warning("Received invalid event type (%X) of %i bytes.",
                    event.type, len(event.data))

    def sysex_event(self, event):
        """Handle system exclusive message events.

        ..note:
            Some MIDI files contain sysex messages which are distributed over
            several events, where the event containing the first part of the
            message has the normal 0xF0 status but the following events have
            a 0xF7 (normally used for escape sequence) and no start-of-sysex,
            i.e. all data bytes. The last part of the message in the sequence
            will have end-of-system-exclusive 0xF7 as the last data byte.

            We assemble these sysex messages here into a single sysex event
            with the timestamp of the event with the first part of the message.

        """
        if self.debug:
            log.debug("System exclusive event, %i bytes", len(event.data) + 1)

        if tointseq(event.data)[-1] == END_OF_EXCLUSIVE:
            if self._sysex_continuation:
                if self.debug:
                    log.debug("Sysex continuation in effect. Appending event data.")
                self._sysex_buffer.data += event.data
                event, self._sysex_buffer = self._sysex_buffer, None

            if self.debug:
                log.debug("Storing sysex mssage of %i bytes", len(event.data) + 1)
            self.add_event(event)
            self._sysex_continuation = False
        else:
            if self.debug:
                log.debug("Sysex message not terminated. Assuming continuation.")
            if self._sysex_continuation:
                self._sysex_buffer.data += event.data
            else:
                self._sysex_buffer = event
            self._sysex_continuation = True

    def escape_sequence(self, event):
        if self.debug:
            log.debug("Escape sequence, value: %r" % event.data)
        self.add_event(event)

    #####################
    ## meta events

    def meta_event(self, event, *args):
        """Dispatch meta events."""
        # TEMPO = 0x51 (51 03 tt tt tt (tempo in us/quarternote))
        track = self.current_track

        if event.meta_type == TEMPO:
            b1, b2, b3 = tointseq(event.data)
            event.tempo = (b1 << 16) + (b2 << 8) + b3
            # uses 3 bytes to represent time between quarter
            # notes in microseconds
            self.tempo(event.tempo)

        # TIME_SIGNATURE = 0x58 (58 04 nn dd cc bb)
        elif event.meta_type == TIME_SIGNATURE:
            nn, dd, cc, bb = tointseq(event.data)
            self.time_signature(nn, dd, cc, bb)

        # KEY_SIGNATURE = 0x59 (59 02 sf mi)
        elif event.meta_type == KEY_SIGNATURE:
            sf, mi = tointseq(event.data)
            self.key_signature(sf, mi)

        # SEQUENCE_NAME = 0x03 (03 len text...)
        elif event.meta_type == SEQUENCE_NAME:
            self.sequence_name(event.data)

        # END_OF_TRACK = 0x2F
        elif event.meta_type == END_OF_TRACK:
            self.end_of_track(track)
        else:
            if self.debug:
                log.debug("Meta event: %s", event)

        self.add_event(event, track)

    def end_of_track(self, track=None):
        """Handle end of current track."""
        self.current_track = None

        if self.debug:
            log.debug("End of track #%i.", track)

    def sequence_name(self, name):
        """Handle Sequence / Track name events."""
        name = name.decode('latin1', 'replace')
        if self.current_track == 0:
            if self.debug:
                log.debug("Sequence name: %s", name)
            self._instance.sequence_name = name
        else:
            if self.debug:
                log.debug("Track name (%02i): %s", self.current_track, name)
            self._instance.tracks[self.current_track] = name

    def tempo(self, value):
        """Handle tempo change events."""
        if not hasattr(self._instance, 'initial_tempo'):
            self._instance.initial_tempo = value

        self.current_tempo = value
        if self.debug:
            log.debug("Tempo: %.2f ms/quarter note (%.2f BPM)",
                      value / 1000.0, 60000000.0 / value)

    def time_signature(self, nn, dd, cc, bb):
        """Handle time signature change events."""
        self.current_time_signature = (nn, dd**2)
        if self.debug:
            log.debug("Time signature: %i/%i (%i cpm, %i dpqn)",
                      nn, dd**2, cc, bb)

    def key_signature(self, sf, mi):
        """Handle key signature change events."""
        if self.debug:
            log.debug("Key signature: %s %s", sf, mi)


if __name__ == '__main__':
    import sys

    if len(sys.argv) > 1:
        test_file = sys.argv[1]
    else:
        print("Usage: sequence.py MIDIFILE")

    logging.basicConfig(level=logging.INFO)
    # do parsing
    sequence = MidiSequence.fromfile(test_file)
    print(sequence)
    sequence.dump_events()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Map note values of all note on/off events according to given mapping file."""

from __future__ import absolute_import, print_function, unicode_literals

import argparse
import csv
import logging
import sys

from os.path import splitext

from miditk.smf import MidiFileReader, MidiFileWriter


log = logging.getLogger("note_remapper")


class NoteRemapper(MidiFileWriter):
    """Map note values of all note on/off events according to given mapping file.
    """
    def __init__(self, fp, mapfn, channel=0, encoding='UTF-8', convert_zero_velocity=False):
        super(NoteRemapper, self).__init__(fp, encoding=encoding)
        self._map = self.read_mapping(mapfn)
        self.channel = channel
        self.convert_zero_velocity = convert_zero_velocity

    def read_mapping(self, mapfn):
        notemap = {}
        with open(mapfn, newline='') as csvfile:
            sample = csvfile.read(1024)
            sniffer = csv.Sniffer()
            dialect = sniffer.sniff(sample)
            csvfile.seek(0)
            reader = csv.reader(csvfile, dialect)

            for lineno, row in enumerate(reader):
                if lineno == 0 and sniffer.has_header(sample):
                    continue

                if len(row) > 2 and row[1].strip():
                    try:
                        srcnote, destnote = int(row[0]), int(row[1])

                        if any((0 > srcnote > 127, 0 > destnote > 127)):
                            raise ValueError("Note number out of range (0 - 127)")
                    except (TypeError, ValueError) as exc:
                        log.warning("Error in mapping file on line %i: %s", lineno, exc)
                    else:
                        notemap[srcnote] = destnote

        return notemap

    def map_note(self, ch, note):
        if self.channel in (0, ch + 1):
            note = max(0, min(127, self._map.get(note, note)))
        return note

    def note_on(self, channel=0, note=0x40, velocity=0x40):
        note = self.map_note(channel, note)
        super(NoteRemapper, self).note_on(channel, note, velocity)

    def note_off(self, channel=0, note=0x40, velocity=0x40):
        note = self.map_note(channel, note)
        super(NoteRemapper, self).note_off(channel, note, velocity)


def main(args=None):
    ap = argparse.ArgumentParser(description=__doc__)
    padd = ap.add_argument
    padd('-v', '--verbose', action="store_true",
         help='verbose output')
    padd('-c', '--channel', metavar="CH", type=int, default=0,
         help='MIDI channel to apply mapping to (default: 0 => all channels)')
    padd(dest='mapfile',
         help='Note mapping file name (CSV format).')
    padd(dest='infile',
         help='Input file name (Standard MIDI File format).')
    padd(dest='outfile', nargs="?",
         help="Output file name (default: add '_remapped' to input file name).")

    args = ap.parse_args(args if args is not None else sys.argv[1:])

    logging.basicConfig(format="%(name)s: %(levelname)s - %(message)s",
                        level=logging.DEBUG if args.verbose else logging.INFO)

    if args.outfile:
        outfile = args.outfile
    else:
        fn, ext = splitext(args.infile)
        outfile = fn + '_remapped' + ext

    # create event handlers
    with open(outfile, 'wb') as smf:
        midiout = NoteRemapper(smf, args.mapfile, channel=args.channel)
        midiin = MidiFileReader(args.infile, midiout)

        # now do the processing
        midiin.read()


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]) or 0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example script which creates a type 0 midi file with a simple drum 4/4 pattern.

All midi events are in the same track chunk on channel 9 (the midi channel
reserved for drums in the General Midi standard).

"""

import sys
from os.path import abspath, dirname, join

from miditk.smf import MidiFileWriter


try:
    outfilename = sys.argv[1]
except IndexError:
    parentdir = dirname(dirname(abspath(__file__)))
    outfilename = join(parentdir, 'tests', 'testoutput', 'drumpattern.mid')

with open(outfilename, 'wb') as smf:
    midi = MidiFileWriter(smf)

    # write file and track header
    midi.header()
    midi.start_of_track()

    # set tempo to 120 bpm
    midi.tempo(int(60000000 / 120))

    # musical events
    # Let's play two bars
    for i in range(4):
        # Acoustic Bass Drum on first/third beat
        midi.note_on(channel=9, note=35)
        # Closed Hi-Hat
        midi.note_on(channel=9, note=42)

        # step one eigth note
        midi.update_ticks(48)
        # Hi-Hat
        midi.note_on(channel=9, note=42)

        # step one eigth note
        midi.update_ticks(48)
        # Acoustic Snare on second/fourth quaver
        midi.note_on(channel=9, note=38)
        # Hi-Hat
        midi.note_on(channel=9, note=42)

        # step one eigth note
        midi.update_ticks(48)
        # Hi-Hat
        midi.note_on(channel=9, note=42)

        # step one eigth note
        midi.update_ticks(48)

    # end track and midi file
    midi.end_of_track()
    midi.eof()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example script which converts a Type 1 SMF to Type 0 file."""

import logging
from operator import attrgetter
from os.path import abspath, dirname, join

from miditk.common.constants import (CUEPOINT, DEVICE_NAME, END_OF_TRACK, INSTRUMENT_NAME,
                                     KEY_SIGNATURE, MARKER, MIDI_PORT, SEQUENCE_NAME,
                                     SEQUENCE_NUMBER, SMTP_OFFSET, TEMPO, TIME_SIGNATURE)
from miditk.smf import MidiFileWriter, MidiSequence


log = logging.getLogger('miditk.smf.examples.type1to0')


# MidiFileWriter is a sub-class of NullMidiEventHandler
class Type0Converter(MidiFileWriter):
    """Write MidiSequence to a Type 0 MIDI file.

    Discards any meta events from track +1..n, which do not make sense in a type 0 file.

    """
    # These meta event types are excluded from the output if they occur on any other than track 0
    ignore_meta_events = (
        CUEPOINT,
        DEVICE_NAME,
        END_OF_TRACK,
        INSTRUMENT_NAME,
        KEY_SIGNATURE,
        MARKER,
        MIDI_PORT,
        SEQUENCE_NAME,
        SEQUENCE_NUMBER,
        SMTP_OFFSET,
        TEMPO,
        TIME_SIGNATURE
    )

    def write(self, sequence):
        log.debug("Writing Type 0 file header.")
        self.header(format=0, num_tracks=1, tick_division=sequence.tick_division,
                    metrical=sequence.metrical, fps=sequence.fps,
                    frame_resolution=sequence.frame_resolution)
        log.debug("Starting track #0.")
        self.start_of_track(0)

        for ticks, events in sequence.events_by_ticks():
            log.debug("Current Tick: %i", ticks)
            self.update_ticks(ticks, relative=False)

            # loop over events for this tick in reverse order of their type code,
            # i.e. meta events will come first, then sysex/escape-sequences, then
            # program/controller changes/pitchbend/aftertouch, then note data.
            for event in sorted(events, key=attrgetter('type'), reverse=True):
                # Ignore all "end of track" meta events,
                # we will add the one for track #0 later.
                if event.meta_type == END_OF_TRACK:
                    continue

                # Ignore meta events from track #0..n, which do not make sense in a Type 0 file.
                if event.track != 0 and event.meta_type in self.ignore_meta_events:
                    log.warning("Discarding event: %r", event)
                    continue

                log.debug("Writing event to track #0: %r", event)
                self.event_slice(event.to_bytes())

        log.debug("Writing end-of-track event to track #0 and writing track.")
        self.end_of_track(0)
        self.eof()


if __name__ == '__main__':
    import sys
    from os.path import splitext

    if '-v' in sys.argv[1:]:
        loglevel = logging.DEBUG
        sys.argv.remove('-v')
    else:
        loglevel = logging.INFO

    logging.basicConfig(level=loglevel)

    parentdir = dirname(dirname(abspath(__file__)))

    if len(sys.argv[1:]) >= 1:
        infilename = sys.argv[1]

        if len(sys.argv[1:]) >= 2:
            outfilename = sys.argv[2]
        else:
            outfilename = splitext(infilename)[0] + '-type0.mid'
    else:
        # Use test file if no input file is given on the command line.
        # This is a small type 1 midi file, with only two midi events on track #1
        # and "sequence name" meta events on all 17 tracks.
        infilename = join(parentdir, 'tests', 'testdata', 'minimal-cubase-type1.mid')
        # The output is written to the tests/testoutput directory
        outfilename = join(parentdir, 'tests', 'testoutput', 'type0-converted.mid')

    # Parse input file to a MidiSequence
    with open(infilename, 'rb') as insmf:
        seq = MidiSequence.fromfile(insmf)

    with open(outfilename, 'wb') as outsmf:
        midiout = Type0Converter(outsmf)
        # now do the processing
        midiout.write(seq)



# -*- coding: utf-8 -*-
#
# miditk/smf/converters.py
#
"""Functions for reading and writing the special data types in a MIDI file."""

__all__ = (
    'read_bew',
    'read_varlen',
    'sizeof_varlen',
    'tointseq',
    'tobytestr',
    'write_bew',
    'write_varlen'
)

# _speedups module not available
from struct import pack, unpack

from six import iterbytes, text_type


if isinstance(b'', str):
    def _tobytes(*values):
        return "".join(chr(c) for c in values)
else:
    def _tobytes(*values):
        return bytes(values)


def _ord(x):
    return x if isinstance(x, int) else ord(x)


def read_bew(value):
    """Read byte string as big endian word.

    Variable Length Data (varlen) is a data format sprayed liberally throughout a midi file. It can
    be anywhere from 1 to 4 bytes long. If the 8'th bit is set in a byte another byte follows. The
    value is stored in the lowest 7 bits of each byte. So maximum value is 2^28 (4x7 bits = 28
    bits) = 268435456.

    ::

        >>> read_bew(b'aáâã')
        1642193635
        >>> read_bew(b'aá')
        25057

    """
    lval = len(value)
    if lval == 1:
        return _ord(value)
    elif lval == 2:
        return unpack('>H', value)[0]
    else:
        return unpack('>L', value)[0]


def read_varlen(value):
    """Convert variable length data format to integer.

    Just pass it 0 or more chars that might be a varlen and it will only use
    the relevant chars.

    Use sizeof_varlen(value) to see how many bytes the integer value takes.

    ::

        >>> read_varlen(b'@')
        64
        >>> read_varlen(b'áâãa')
        205042145

    """
    sum = 0
    for byte in iterbytes(value):
        sum = (sum << 7) + (byte & 0x7F)

        if not 0x80 & byte:
            # stop after last byte with high bit set
            break

    return sum


def sizeof_varlen(value):
    """Return number of bytes an integer will need when converted to varlength."""
    if value <= 127:
        return 1
    elif value <= 16383:
        return 2
    elif value <= 2097151:
        return 3
    else:
        return 4


def tobytestr(value, encoding='latin1'):
    """Convert given string or sequence of integers to a byte string."""
    if isinstance(value, text_type):
        value = value.encode(encoding)
    elif isinstance(value, (list, tuple)):
        value = _tobytes(*value)
    return value


def tointseq(value, encoding='latin1'):
    """Convert a bytes/str/unicode instance into a tuple of int byte values."""
    if isinstance(value, text_type):
        value = value.encode(encoding)
    return tuple(_ord(c) for c in value)


def write_bew(value, length):
    """Write int as big endian formatted bytes string.

    >>> read_bew(write_bew(25057, 2))
    25057
    >>> long(read_bew(write_bew(1642193635L, 4)))
    1642193635L

    """
    if length == 1:
        return _tobytes(value)
    elif length == 2:
        return pack('>H', value)
    else:
        return pack('>L', value)


def write_varlen(value):
    """Convert an integer to varlength format."""
    nbytes = sizeof_varlen(value)
    if nbytes == 1:
        return _tobytes(value)
    else:
        return b"".join(_tobytes(((value >> ((i - 1) * 7)) | 0x80) & (0xFF if i > 1 else 0x7F))
                        for i in range(nbytes, 0, -1))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example script which maps program change numbers."""

from miditk.smf import MidiFileReader, MidiFileWriter


class ProgramChangeMapper(MidiFileWriter):
    """Change program number of Program Change events according to given mapping."""

    def __init__(self, outfile, pcmap, channel=None, track=None):
        self.pcmap = pcmap
        self.channel = (channel,) if isinstance(channel, int) else channel
        self.track = (track,) if isinstance(track, int) else track
        super().__init__(outfile)

    def program_change(self, channel, program):
        if ((self.channel is None or channel in self.channel) and
                (self.track is None or self.current_track in self.track)):
            program = self.pcmap.get(program, program)
        super().program_change(self, channel, min(127, max(0, program)))


if __name__ == '__main__':
    import sys
    infilename = sys.argv.pop(1)
    outfilename = sys.argv.pop(1)
    # Program change mapping
    pcmap = {
        0: 40,   # Ac. Piano -> Strings
        32: 33,  # Ac. Bass -> Electric Bass
    }
    # Create parser and event handler
    with open(outfilename, 'wb') as smf:
        midiout = ProgramChangeMapper(smf, pcmap)
        midiin = MidiFileReader(infilename, handler=midiout)
        # Now do the processing.
        midiin.read()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# midi2syx.py
#
"""Extract all system excusive messages from a standard MIDI file.

By default writes each sysex event contained in the MIDI file to its own file
in the output directory (defaults to the current working directory) which
contains the pure sysex data. The filename of each sysex file is constructed
by appending the number of the sysex event to a common prefix, which can be
specified via an option (defaults to `msg_`).

"""

from __future__ import print_function

import logging
import argparse
import os
import sys

from miditk.common.constants import SYSTEM_EXCLUSIVE
from miditk.smf import MidiSequence


__version__ = "0.2"
__usage__ = "Usage: mid2syx.py MIDIFILE"


def _ord(x):
    return x if isinstance(x, int) else ord(x)


def error(msg, *args):
    print(msg % args, stream=sys.stderr)


def checksum(msg, data_offset):
    return sum(_ord(c) for c in msg[data_offset:-2]) & 0x7f


def sysex_info(msg, msg_no):
    if _ord(msg[0]) == 0:
        manufacturer_id = '%Xh %Xh' % (_ord(msg[1]), _ord(msg[2]))
        model_id = _ord(msg[3])
        device_id = _ord(msg[4])
    else:
        manufacturer_id = '%Xh' % _ord(msg[0])
        model_id = _ord(msg[2])
        device_id = _ord(msg[3])

    return ("Msg #%03i, %i bytes, manufacturer: %s, model: %03Xh, device ID %03Xh" %
            (msg_no, len(msg) + 1, manufacturer_id, model_id, device_id))


def main(args=None):
    ap = argparse.ArgumentParser(usage=__usage__, description=__doc__)
    aa = ap.add_argument
    aa('-c', '--checksum', action="store_true", help="Verify check sum of each sysex message.")
    aa('-n', '--dry-run', action="store_true",
       help="Only print what would be done, do not write any files.")
    aa('-p', '--prefix', metavar="PREFIX", default="msg_",
       help="Filename prefix for each sysex file.")
    aa('-o', '--output-dir', metavar="DIR", default=os.getcwd(),
       help="Output directory where extracted sysex files will be written.")
    aa('-q', '--quiet', action="store_true",
       help="Do not print anything to standard output at all.")
    aa('-v', '--verbose', action="store_true",
       help="Print debugging info to standard output.")
    aa('midifile', metavar="MIDIFILE",
       help="Standard MIDI input file.")
    aa('--version', action='version', version="mid2sys " + __version__)

    args = ap.parse_args(args if args is not None else sys.argv[1:])

    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.WARNING)

    if not os.path.isdir(args.output_dir):
        error("Not a directory: '%s'\n", args.output_dir)
        ap.print_help()
        return 1

    sequence = MidiSequence.fromfile(args.midifile)

    for i, ev in enumerate(sequence.sysex_events()):
        assert ev.type == SYSTEM_EXCLUSIVE
        assert _ord(ev.data[0]) < 127
        assert ev.data.endswith(b'\xF7')

        if not args.quiet:
            print("Track #%02i: %s" % (ev.track + 1, sysex_info(ev.data, i + 1)))

        if args.checksum:
            cs = checksum(ev.data, 6)
            if cs != _ord(ev.data[-2]):
                error("Checksum verification for msg #%03i failed! "
                      "Expected %i, read %i", i + 1, cs, _ord(ev.data[-2]))

        outfile = os.path.join(args.output_dir, "%s%03i.syx" % (args.prefix, i + 1))

        try:
            if not args.dry_run:
                fp = open(outfile.encode('utf-8'), 'wb')
        except (IOError, OSError) as exc:
            error("Could not open output file '%s' for writing: %s", outfile, exc)
        else:
            if args.verbose:
                print("Writing sysex msg #%03i to '%s'..." % (i + 1, outfile))
            if not args.dry_run:
                fp.write(b"\xF0" + ev.data)
                fp.close()

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]) or 0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# profile_miditoobject.py
#

from __future__ import absolute_import, print_function, unicode_literals

import sys
import cProfile

from miditk.smf.sequence import MidiSequence  # noqa:F401


if len(sys.argv) < 2:
    print("Usage: profile_miditoobject.py MIDIFILE")
else:
    filename = sys.argv.pop(1)

    if len(sys.argv) > 1:
        cProfile.run("MidiSequence.fromfile(filename)", sys.argv[1], sort='tottime')
    else:
        cProfile.run("MidiSequence.fromfile(filename)", sort='tottime')



# -*- coding: utf-8 -*-
#
# miditk/smf/parser.py
#
"""MIDI file low-level binary parsing."""

from __future__ import absolute_import, unicode_literals

# standard library imports
import logging

# package-specific imports
from ..common.constants import (CHANNEL_PRESSURE, CONTROLLER_CHANGE, END_OF_EXCLUSIVE,
                                ESCAPE_SEQUENCE, META_EVENT, MTC, NOTE_OFF, NOTE_ON, PITCH_BEND,
                                POLY_PRESSURE, PROGRAM_CHANGE, SONG_POSITION_POINTER, SONG_SELECT,
                                SYSTEM_EXCLUSIVE)
from .api import MidiEvent
from .converters import read_bew, read_varlen, sizeof_varlen, tointseq

__all__ = ('MidiFileParser', 'ParseError')
log = logging.getLogger(__name__)


# exceptions
class ParseError(Exception):
    pass


# utility functions
def _read_event_data(stream):
    """Read data for next MIDI event from stream.

    Returns a tuple with (size, data).

    """
    data = stream.read(4)
    size = read_varlen(data)
    stream.seek(sizeof_varlen(size) - len(data), 1)
    return size, stream.read(size)


# classes
class MidiFileParser(object):
    """Parser that decodes the raw binary midi data.

    The MidiFileParser is the lowest level parser and generates events from
    the raw midi data, which are processed by the output event handlers.

    The idea is that this class does not retain any status other than is needed
    to parse the raw midi data. Instead it calls methods on the handler object
    whenever an event occurs or the status changes (i.e. parsing events or
    event timings). It is then the responsibility of the handler to keep track
    of status.

    """

    channel_data_sizes = {
        CHANNEL_PRESSURE: 1,
        CONTROLLER_CHANGE: 2,
        NOTE_OFF: 2,
        NOTE_ON: 2,
        PITCH_BEND: 2,
        POLY_PRESSURE: 2,
        PROGRAM_CHANGE: 1,
    }

    system_data_sizes = {
        MTC: 1,
        SONG_POSITION_POINTER: 2,
        SONG_SELECT: 1
    }

    def __init__(self, instream, handler=None, strict=True):
        """Initialize object.

        instream must be an open, readable file-like object for a standard MIDI file.

        The file-like object must support the ``read``, ``seek`` and ``tell` methods.

        """
        self.instream = instream
        self._handler = handler
        self.strict = strict
        self._current_track = None
        self._running_status = None
        self._sysex_continuation = False

    def dispatch(self, methodname, *args, **kwargs):
        try:
            try:
                method = getattr(self._handler, methodname)
            except AttributeError:
                msg = "Method '%s' not implemented by handler." % methodname
                if self.strict:
                    raise AttributeError(msg)
                else:
                    log.warning(msg)
            else:
                method(*args, **kwargs)
        except Exception as exc:
            msg = "Error calling handler '%s'" % methodname

            if args and isinstance(args[0], MidiEvent):
                msg += " for event %r" % args[0]
            else:
                msg += "Args: %r, Kw args: %r " % (args, kwargs)

            if self.strict:
                log.exception(msg)
                raise
            else:
                log.error(msg + str(exc))

    def next_chunk(self):
        chunk_id = self.instream.read(4)

        if not chunk_id:
            raise EOFError("End of file reached.")

        chunk_len = read_bew(self.instream.read(4))
        return chunk_id, chunk_len

    def parse_header(self):
        """Parse the header chunk."""

        chunk_id, chunk_len = self.next_chunk()
        instream = self.instream

        # check if it is a proper midi file
        if chunk_id != b'MThd':
            raise ParseError("Invalid MIDI file header. Chunk identifier must be 'MThd'.")

        # Header values are at fixed locations, so no reason to be clever
        self.format = read_bew(instream.read(2))
        self.num_tracks = read_bew(instream.read(2))

        if self.format == 0 and self.num_tracks > 1:
            msg = ("Invalid number of tracks (%i). Type 0 midi files may only "
                   "contain a single track." % self.num_tracks)

            if self.strict:
                raise ParseError(msg)
            else:
                log.warning(msg)

        tick_div = instream.read(2)
        fps, resolution = tointseq(tick_div)

        if fps & 0x80:
            metrical = False
        else:
            metrical = True
            division = read_bew(tick_div)

        # Theoretically a header larger than 6 bytes can exist
        # but no one has seen one in the wild.
        # We will correctly ignore unknown data if present, though.
        if chunk_len > 6:
            log.warning("Invalid header size (%i). Skipping trailing header "
                        "bytes", chunk_len)
            instream.seek(chunk_len - 6, 1)

        # call the header event handler on the stream
        if metrical:
            self.dispatch('header', self.format, self.num_tracks, metrical=True,
                          tick_division=division)
        else:
            self.dispatch('header', self.format, self.num_tracks, metrical=False,
                          fps=fps, frame_resolution=resolution)

    def parse_track(self, chunk_id, chunk_len):
        """Parse a track chunk.

        This is the most important part of the parser.

        """
        self._running_status = None
        self._sysex_continuation = False

        instream = self.instream
        dispatch = self.dispatch

        # check for proper chunk type
        if chunk_id != b'MTrk':
            raise ParseError("Invalid track chunk identifier '%s', must be 'MTrk'.", chunk_id)

        if self._current_track is None:
            self._current_track = 0
        else:
            self._current_track += 1

        if self._current_track >= self.num_tracks:
            msg = ("Supernumerous track no. %i found. Header says there should "
                   "be only %i tracks." % (self._current_track + 1, self.num_tracks))

            if self.strict:
                raise ParseError(msg)
            else:
                log.warning(msg)

        # Trigger event at the start of a track
        # set time to 0 at start of a track
        dispatch('reset_ticks')
        dispatch('start_of_track', self._current_track)
        # absolute position!
        track_endposition = instream.tell() + chunk_len

        while instream.tell() < track_endposition:
            # find relative time of the event
            ticks = read_varlen(instream.read(4))
            instream.seek(sizeof_varlen(ticks) - 4, 1)

            if ticks > 0:
                dispatch('update_ticks', ticks)

            # be aware of running status!
            status = ord(instream.read(1))

            if status & 0x80:
                # the status byte has the high bit set, so it
                # was not running data but proper status byte
                status = self._running_status = status
            else:
                # use saved running status
                if self._running_status:
                    status = self._running_status
                    instream.seek(-1, 1)
                else:
                    msg = ("Non-status byte 0x%02X encountered at offset %i "
                           "while expecting a status byte and no running status "
                           "in effect.")
                    offset = instream.tell()

                    if self.strict:
                        log.error(msg, status, offset)
                        raise ParseError(msg % (status, offset))
                    else:
                        log.warning(msg, status, offset)
                        log.warning("Trying to re-synchronize.")
                        continue

            # Create event container
            event = MidiEvent(track=self._current_track)

            # Is it a meta event?
            # these only exists in midi files, not in transmitted midi data.
            # In transmitted data META_EVENT (0xFF) is a system reset.
            if status == META_EVENT:
                self._running_status = None
                event.meta_type = ord(instream.read(1))
                event.size, event.data = _read_event_data(instream)
                dispatch('meta_event', event)

            # Is it a sysex event?
            elif status == SYSTEM_EXCLUSIVE:
                self._running_status = None
                event.size, event.data = _read_event_data(instream)
                event.type = SYSTEM_EXCLUSIVE

                # We check for proper terminiation of the sysex msg
                # with an EOX byte, since some manufacturers split up sysex
                # msgs over several events (continuation event).
                if event.data[-1] != END_OF_EXCLUSIVE:
                    self._sysex_continuation = True
                else:
                    self._sysex_continuation = False

                dispatch('sysex_event', event)

            # Escape sequence or sysex continuation message
            elif status == END_OF_EXCLUSIVE:
                self._running_status = None

                event.size, event.data = _read_event_data(instream)
                if self._sysex_continuation:
                    event.type = SYSTEM_EXCLUSIVE

                    if event.data[-1] == END_OF_EXCLUSIVE:
                        self._sysex_continuation = False

                    dispatch('sysex_event', event)
                else:
                    event.type = ESCAPE_SEQUENCE
                    dispatch('escape_sequence', event)

            elif 0xF1 <= status <= 0xFE:
                # Invalid system common or system real time message
                self._running_status = None
                event.data_size = self.system_data_sizes.get(status, 0)
                event.data = instream.read(event.data_size)
                event.type = status
                dispatch('invalid_event', event)

            # Otherwise it must be a channel voice message
            else:
                event.data_size = self.channel_data_sizes.get(status & 0xF0, 0)
                event.data = instream.read(event.data_size)
                event.type = status & 0xF0
                event.channel = status & 0xF
                dispatch('channel_message_event', event)

    def parse_tracks(self):
        """Parse all track chunks."""
        while True:
            try:
                chunk_id, chunk_len = self.next_chunk()
            except EOFError:
                break
            else:
                if chunk_id == b'MTrk':
                    self.parse_track(chunk_id, chunk_len)
                else:
                    self.instream.seek(chunk_len, 1)

        self.dispatch('eof')


if __name__ == '__main__':
    from miditk.smf.api import PrintingMidiEventHandler

    # get data
    test_files = (
        'tests/testdata/minimal.mid',
        'tests/testdata/minimal-cubase-type0.mid',
        'tests/testdata/minimal-cubase-type1.mid',
    )

    # do parsing
    for test_file in test_files:
        with open(test_file, 'rb') as midi_file:
            midi_in = MidiFileParser(midi_file, PrintingMidiEventHandler())
            midi_in.parse_header()
            midi_in.parse_tracks()
            print('')
